// https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#high-level-structure
#define WASM_VERSION 0x1

// Section codes
#define SC_TYPE     1
#define SC_IMPORT   2
#define SC_FUNCTION 3
#define SC_TABLE    4
#define SC_MEMORY   5
#define SC_GLOBAL   6
#define SC_EXPORT   7
#define SC_START    8
#define SC_ELEMENT  9
#define SC_CODE     10
#define SC_DATA     11

// https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#language-types
#define TYPE_I32        0x7f
#define TYPE_I64        0x7e
#define TYPE_F32        0x7d
#define TYPE_F64        0x7c
#define TYPE_ANYFUNC    0x70
#define TYPE_FUNC       0x60
#define TYPE_VOID_BLOCK 0x40

// https://github.com/WebAssembly/design/blob/master/BinaryEncoding.md#external_kind
#define EXT_FUNCTION    0
#define EXT_TABLE       1
#define EXT_MEMORY      2
#define EXT_GLOBAL      3

// Taken from wasm2ast's source code and modified with vim magic
#define OP_UNREACHABLE          0x00
#define OP_NOP                  0x01
#define OP_BLOCK                0x02
#define OP_LOOP                 0x03
#define OP_IF                   0x04
#define OP_ELSE                 0x05
#define OP_TRY                  0x06
#define OP_CATCH                0x07
#define OP_THROW                0x08
#define OP_RETHROW              0x09
#define OP_CATCHALL             0x0a
#define OP_END                  0x0b
#define OP_BR                   0x0c
#define OP_BRIF                 0x0d
#define OP_BRTABLE              0x0e
#define OP_RETURN               0x0f
#define OP_CALL                 0x10
#define OP_CALLINDIRECT         0x11
#define OP_DROP                 0x1a
#define OP_SELECT               0x1b
#define OP_GETLOCAL             0x20
#define OP_SETLOCAL             0x21
#define OP_TEELOCAL             0x22
#define OP_GETGLOBAL            0x23
#define OP_SETGLOBAL            0x24
#define OP_I32LOAD              0x28
#define OP_I64LOAD              0x29
#define OP_F32LOAD              0x2a
#define OP_F64LOAD              0x2b
#define OP_I32LOAD8S            0x2c
#define OP_I32LOAD8U            0x2d
#define OP_I32LOAD16S           0x2e
#define OP_I32LOAD16U           0x2f
#define OP_I64LOAD8S            0x30
#define OP_I64LOAD8U            0x31
#define OP_I64LOAD16S           0x32
#define OP_I64LOAD16U           0x33
#define OP_I64LOAD32S           0x34
#define OP_I64LOAD32U           0x35
#define OP_I32STORE             0x36
#define OP_I64STORE             0x37
#define OP_F32STORE             0x38
#define OP_F64STORE             0x39
#define OP_I32STORE8            0x3a
#define OP_I32STORE16           0x3b
#define OP_I64STORE8            0x3c
#define OP_I64STORE16           0x3d
#define OP_I64STORE32           0x3e
#define OP_CURRENTMEMORY        0x3f
#define OP_GROWMEMORY           0x40
#define OP_I32CONST             0x41
#define OP_I64CONST             0x42
#define OP_F32CONST             0x43
#define OP_F64CONST             0x44
#define OP_I32EQZ               0x45
#define OP_I32EQ                0x46
#define OP_I32NE                0x47
#define OP_I32LTS               0x48
#define OP_I32LTU               0x49
#define OP_I32GTS               0x4a
#define OP_I32GTU               0x4b
#define OP_I32LES               0x4c
#define OP_I32LEU               0x4d
#define OP_I32GES               0x4e
#define OP_I32GEU               0x4f
#define OP_I64EQZ               0x50
#define OP_I64EQ                0x51
#define OP_I64NE                0x52
#define OP_I64LTS               0x53
#define OP_I64LTU               0x54
#define OP_I64GTS               0x55
#define OP_I64GTU               0x56
#define OP_I64LES               0x57
#define OP_I64LEU               0x58
#define OP_I64GES               0x59
#define OP_I64GEU               0x5a
#define OP_F32EQ                0x5b
#define OP_F32NE                0x5c
#define OP_F32LT                0x5d
#define OP_F32GT                0x5e
#define OP_F32LE                0x5f
#define OP_F32GE                0x60
#define OP_F64EQ                0x61
#define OP_F64NE                0x62
#define OP_F64LT                0x63
#define OP_F64GT                0x64
#define OP_F64LE                0x65
#define OP_F64GE                0x66
#define OP_I32CLZ               0x67
#define OP_I32CTZ               0x68
#define OP_I32POPCNT            0x69
#define OP_I32ADD               0x6a
#define OP_I32SUB               0x6b
#define OP_I32MUL               0x6c
#define OP_I32DIVS              0x6d
#define OP_I32DIVU              0x6e
#define OP_I32REMS              0x6f
#define OP_I32REMU              0x70
#define OP_I32AND               0x71
#define OP_I32OR                0x72
#define OP_I32XOR               0x73
#define OP_I32SHL               0x74
#define OP_I32SHRS              0x75
#define OP_I32SHRU              0x76
#define OP_I32ROTL              0x77
#define OP_I32ROTR              0x78
#define OP_I64CLZ               0x79
#define OP_I64CTZ               0x7a
#define OP_I64POPCNT            0x7b
#define OP_I64ADD               0x7c
#define OP_I64SUB               0x7d
#define OP_I64MUL               0x7e
#define OP_I64DIVS              0x7f
#define OP_I64DIVU              0x80
#define OP_I64REMS              0x81
#define OP_I64REMU              0x82
#define OP_I64AND               0x83
#define OP_I64OR                0x84
#define OP_I64XOR               0x85
#define OP_I64SHL               0x86
#define OP_I64SHRS              0x87
#define OP_I64SHRU              0x88
#define OP_I64ROTL              0x89
#define OP_I64ROTR              0x8a
#define OP_F32ABS               0x8b
#define OP_F32NEG               0x8c
#define OP_F32CEIL              0x8d
#define OP_F32FLOOR             0x8e
#define OP_F32TRUNC             0x8f
#define OP_F32NEAREST           0x90
#define OP_F32SQRT              0x91
#define OP_F32ADD               0x92
#define OP_F32SUB               0x93
#define OP_F32MUL               0x94
#define OP_F32DIV               0x95
#define OP_F32MIN               0x96
#define OP_F32MAX               0x97
#define OP_F32COPYSIGN          0x98
#define OP_F64ABS               0x99
#define OP_F64NEG               0x9a
#define OP_F64CEIL              0x9b
#define OP_F64FLOOR             0x9c
#define OP_F64TRUNC             0x9d
#define OP_F64NEAREST           0x9e
#define OP_F64SQRT              0x9f
#define OP_F64ADD               0xa0
#define OP_F64SUB               0xa1
#define OP_F64MUL               0xa2
#define OP_F64DIV               0xa3
#define OP_F64MIN               0xa4
#define OP_F64MAX               0xa5
#define OP_F64COPYSIGN          0xa6
#define OP_I32WRAPI64           0xa7
#define OP_I32TRUNCSF32         0xa8
#define OP_I32TRUNCUF32         0xa9
#define OP_I32TRUNCSF64         0xaa
#define OP_I32TRUNCUF64         0xab
#define OP_I64EXTENDSI32        0xac
#define OP_I64EXTENDUI32        0xad
#define OP_I64TRUNCSF32         0xae
#define OP_I64TRUNCUF32         0xaf
#define OP_I64TRUNCSF64         0xb0
#define OP_I64TRUNCUF64         0xb1
#define OP_F32CONVERTSI32       0xb2
#define OP_F32CONVERTUI32       0xb3
#define OP_F32CONVERTSI64       0xb4
#define OP_F32CONVERTUI64       0xb5
#define OP_F32DEMOTEF64         0xb6
#define OP_F64CONVERTSI32       0xb7
#define OP_F64CONVERTUI32       0xb8
#define OP_F64CONVERTSI64       0xb9
#define OP_F64CONVERTUI64       0xba
#define OP_F64PROMOTEF32        0xbb
#define OP_I32REINTERPRETF32    0xbc
#define OP_I64REINTERPRETF64    0xbd
#define OP_F32REINTERPRETI32    0xbe
#define OP_F64REINTERPRETI64    0xbf

#define MEM_NO_ALIGN 0
#define MEM_ALIGN16 1
#define MEM_ALIGN32 2
#define MEM_IMM_OFFSET 0
